"""
This module contains some constants used throughout the package.
"""

#: The node attribute used to store the label of a node.
LABEL_ATTR = "label"
#: The edge attribute used to store the length of an edge.
LENGTH_ATTR = "length"
#: Prefix for the node name of a reticulation node when parsing a phylogenetic network from newick
RETIC_PREFIX = "__#R"
