"""
A module for constructing and reducing networks using cherry-picking.

This code was initially based on a collection of scripts used for several papers:
 - "On cherry-picking and network containment" by Remie Janssen and Yukihiro Murakami (2020)
 - "Combining networks using cherry picking sequences" by Remie Janssen, Mark Jones, and Yukihiro Murakami (2020)
 - "Linear Time Algorithm for Tree-Child network Containment" by Remie Janssen and Yukihiro Murakami (2020)
 - "A practical fixed-parameter algorithm for constructing tree-child networks from multiple binary trees" by Leo van Iersel, Remie Janssen, Mark Jones, Yukihiro Murakami, and Norbert Zeh (2022)
and some unpublished work that was continued by Esther Julien:
 - Constructing Phylogenetic Networks via Cherry Picking and Machine Learning" by G BErnardini, LJJ van Iersel, E Julien, and L Stougie (2023)
"""

from .base import *
