"""
Heuristic distance algorithms for rearrangement distances.

This module contains the heuristic distance algorithms for rearrangement distances.
The algorithms are based on thescripts used for PhD thesis of R. Janssen, 
"Rearranging Phylogenetic Networks", 2021.

The proofs originate from constructive proofs by induction for 
bounds on the distance between two networks. These proofs can be found in
the following papers:
 - Exploring the tiers of rooted phylogenetic network space using tail moves
R Janssen, M Jones, PL Erdős, L van Iersel, C Scornavacca
Bulletin of mathematical biology 80 (8), 2177-2208
 - Heading in the right direction? Using head moves to traverse phylogenetic network space
R Janssen
Journal of Graph Algorithms and Applications 25 (1), 263-310
 - Rearranging phylogenetic networks
R Janssen
PhD thesis
"""
